from . import audio
from . import curvefit
from . import data
from . import dataset
from . import entry_points
from . import general
from . import labels
from . import path
from . import splitalgos
from . import spect
from . import validation

from .split import train_test_dur_split
